import numpy as np
from scipy.special import erf
from iminuit import Minuit


def histogram_errors(bin_counts):
    """
    Calculate the errors for histogram bins.

    This function computes the error for each bin in a histogram,
    treating positive and negative bin counts differently.

    Args:
        x (np.ndarray): Input array of histogram bin counts.

    Returns:
        np.ndarray: Array of calculated errors.
    """
    out = np.ones_like(bin_counts)

    # Handle positive values
    positive_mask = bin_counts > 0
    out[positive_mask] = np.sqrt(bin_counts[positive_mask])

    # Handle negative values
    negative_mask = bin_counts < 0
    out[negative_mask] = np.sqrt(-bin_counts[negative_mask])
    return out


def chunk_multi(data, block_size):
    """
    Chunks input array along the 'row' and 'col' dimensions.

    Args:
        data (ndarray): input data array, at least 2D.
        block_size (list or tuple of int): [chunk_row, chunk_col],
        dimension of the 2-D chunk.

    Returns: list of chunked data.

    Raises:
        ValueError: If input data is not at least 2D or block_size is invalid.
    """
    if not isinstance(data, np.ndarray):
        raise ValueError("Input data must be a numpy array.")
    if data.ndim < 2:
        raise ValueError("Input data must be at least 2D.")
    if (
        not isinstance(block_size, (list, tuple)) or
        len(block_size) != 2 or
        (not all(isinstance(x, int) and x > 0 for x in block_size))
    ):
        raise ValueError(
            "block_size must be a list or tuple of "
            "two positive integers.")

    rows, cols = data.shape[-2:]
    chunk_rows, chunk_cols = block_size

    if chunk_rows > rows or chunk_cols > cols:
        raise ValueError(
            "block_size dimensions cannot be larger "
            "than the input array dimensions.")

    chunks = []
    for i in range(0, rows, chunk_rows):
        for j in range(0, cols, chunk_cols):
            chunk = data[
                ...,
                i:min(i+chunk_rows, rows),
                j:min(j+chunk_cols, cols)
            ]
            chunks.append(chunk)

    return chunks


def fill_histogram(image_data, histogram_bins):
    """
    Fill a histogram from input images.
    This function creates a histogram with shape
    (n_bins-1, memory_cells, n_rows, n_cols) from the input image data.

    Args:
        histogram_bins (list, float): the binning of the x-axis
        image_data (np.ndarray): image data to histogram
            (trains, memory_cells, row, col).

    Returns: histogram bin counts, bin edges.

    Raises
        TypeError: If imgs is not a numpy ndarray.
        ValueError: If imgs is not a 4D array.
    """
    if not isinstance(image_data, np.ndarray):
        raise TypeError("Expected imgs numpy ndarray type.")

    if image_data.ndim < 4:
        raise ValueError("Expected 4D imgs array.")

    n_cells, n_rows, n_cols = image_data.shape[1:]

    h_chk = np.zeros(
        (len(histogram_bins)-1, n_cells, n_rows, n_cols),
        dtype=np.int32)

    e_chk = None
    for cell in range(n_cells):
        for row in range(n_rows):
            for col in range(n_cols):
                this_pix = image_data[:, cell, row, col]
                _h, _e = np.histogram(this_pix, bins=histogram_bins)
                h_chk[..., cell, row, col] += _h
                if e_chk is None:
                    e_chk = np.array(_e)
    return h_chk, e_chk


# peak finder to find the first photon peak and/or the pedestal peak
# can/should be replaced with smthg more efficient

def _peak_position(x, h, thr=75, dist=30., ratio=0.4):
    """
    finds peaks in the histogram
    very rough, self-made function, should be replaced by better one

    arguments:
    * x (list, float): histogram bin centers
    * h (list, float): histogram bin counts
    * thr (int): lower threshold along x-axis in ADC units to look for a peak
    * dist (float): minimal distance between the last found peak and the next
    * ratio (float): minimal ratio between the last found peak and the next

    returns: list of sorted peak positions along the x-axis, index
    of the peak with lowes value along x-axis
    """
    imin = 0
    peak_bins = []
    if len(x) == len(h):
        for i in range(0, len(x)):
            if x[i] > thr:
                imin = i
                break

        sorted_i = np.argsort(h[imin:])
        peak_bins.append(sorted_i[-1] + imin)
        low_h = ratio * h[sorted_i[-1] + imin]
        for j in range(1, len(sorted_i)):
            ibin = sorted_i[-1 - j] + imin
            if h[ibin] > low_h:
                min_dist = (x[peak_bins] - x[ibin])**2. > dist**2.
                if np.all(min_dist):
                    peak_bins.append(ibin)
    else:
        raise AttributeError(
            'axis and histogram arrays have '
            f'different length: {len(x)} and {len(h)}')

    return peak_bins, imin


def set_histo_range(bin_centers, histogram, h_range):
    """
    Reduces the histogram bin axis to the range specified

    Args:
        bin_centers (array, float): the bin centers array
        histogram (array, integers): the histogram with shape
            (bins, cells, columns, row)
        h_range (list, float): the (liminf, limsup) of the desired range

    Returns: the new bin centers array and the new histogram

    """
    if len(h_range) != 2:
        raise ValueError("h_ranges should have two values.")
    min_val, max_val = sorted(h_range)

    i_min = np.abs(bin_centers - min_val).argmin()
    i_max = np.abs(bin_centers - max_val).argmin()

    # Ensure i_max is greater than i_min
    if i_max <= i_min:
        i_max = i_min + 1

    return bin_centers[i_min:i_max], histogram[i_min:i_max, ...]


def rebin_histo(hist, bin_centers, rebin_factor):
    """
    Rebin a histogram by combining adjacent bins.

    Args:
        hist (np.ndarray): Input histogram values.
        bin_edges (np.ndarray): Bin edges or centers.
        rebin_factor (int): Rebinning factor. Number of adjacent
            bins to combine.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Rebinned histogram values and
            corresponding bin edges/centers.

    Raises:
        ValueError: If the rebinning factor is not an exact divisor
            of the number of bins.
    """
    if rebin_factor == 1:
        return hist, bin_centers

    # Check if rebinning is possible
    if len(hist) % rebin_factor != 0:
        raise ValueError(
            "Rebin factor is not an exact divisor "
            "of the number of bins.")

    h_out = None
    x_out = bin_centers[::rebin_factor]

    h_out = np.sum(
        hist[:len(hist) - (len(hist) % rebin_factor)].reshape(-1, rebin_factor),  # noqa
        axis=1,
    )
    x_out = bin_centers[::rebin_factor]
    return h_out, x_out


def fit_histogram(
    bin_centers,
    fit_func,
    n_sigma,
    rebin,
    ratio,
    noise,
    initial_sigma,
    histo,  # Added at the end to parallelize with multiprocessing.
):
    """
    Wrap around function for fitting of histogram

    Args:
        bin_centers (list, float): - bin centers along x
        fit_func (string): which function to use for fit
            - CHARGE_SHARING: single peak with charge sharing tail
            - CHARGE_SHARING_2: sum of two CHARGE_SHARING
            - GAUSS: gaussian function
        n_sigma (int): to calculate threshold of the peak finder as
            thr = n_sigma * initial_sigma
        ratio (float): ratio parameter of the peak finder
        histo (ndarray): histogram bin counts with shape
            (n_bins, memory_cells, row, col)

    Returns:
        - map of peak values
        - map of peak variances
        - map of charge sharing parameter values
        - map of chi2/ndf
    """
    _funcs = dict(
        CHARGE_SHARING=fit_charge_sharing,
        CHARGE_SHARING_2=fit_double_charge_sharing,
        GAUSS=fit_gauss,
    )
    fit_func = _funcs[fit_func]
    n_cells, n_rows, n_cols = histo.shape[1:]

    _init_array = np.zeros((n_cells, n_rows, n_cols), dtype=np.float32)
    peak_positions = _init_array.copy()
    sigma_chk = _init_array.copy()
    chi2ndf_chk = _init_array.copy()
    alpha_chk = _init_array.copy()

    for cell in range(n_cells):
        for row in range(n_rows):
            for col in range(n_cols):
                y = histo[:, cell, row, col]
                y, x_rebinned = rebin_histo(y, bin_centers, rebin)
                yerr = histogram_errors(y)

                if noise[0, cell, row, col] > 0.:
                    initial_sigma = noise[0, cell, row, col]

                q, sigma, chi2ndf, alpha = fit_func(
                    x_rebinned, y, yerr, initial_sigma, n_sigma, ratio)

                peak_positions[cell, row, col] = q
                sigma_chk[cell, row, col] = sigma
                chi2ndf_chk[cell, row, col] = chi2ndf
                alpha_chk[cell, row, col] = alpha
    return peak_positions, sigma_chk, chi2ndf_chk, alpha_chk


def erfbox(x, ped, q, sigma):
    """
    Compute the error function box used in charge sharing calculations.

    This function models the box-like shape of charge collection efficiency.

    erfbox is mostly flat between ped and q0 and quickly goes to zero
    outside the range [ped, q0]. The function integral is normalized to 1.
    The slope of the box sides is given by erf(x)/sigma; if sigma == 0 then the
    function goes abruptly to 0 at the edges.

    Args:
        x (array-like): Input values
        ped (float): Pedestal value
        q0 (float): Charge value
        sigma (float): Width parameter

    Returns:
        array-like: Computed erfbox values
    """
    if ped == q:
        return np.zeros_like(x)
    if ped > q:
        q, ped = ped, q

    slope = lambda x: erf(x / abs(sigma)) if sigma != 0 else 2 * np.heaviside(x, 1)  # noqa
    return (slope(x - ped) - slope(x - q)) * 0.5 / (q - ped)


def gauss(x, amp, mean, sigma):
    """
    Compute a Gaussian function.

    This function calculates the values of a Gaussian distribution
    for given input parameters.
    """
    z = (x - mean) / sigma
    return amp * np.exp(-0.5 * z**2)


def charge_sharing(x, ped, q, sigma, alpha, norm):
    """
    Model the detector response to a single photon with charge sharing.

    This function combines a Gaussian peak
    (representing full charge collection) with a charge sharing tail.

    Args:
        x (array-like): Input charge values
        ped (float): Pedestal value (usually 0 for pedestal-subtracted data)
        q (float): Peak position (full charge collection)
        sigma (float): Peak width
        alpha (float): Charge sharing probability
        norm (float): Normalization factor

    Returns:
        array-like: Modeled detector response

    Notes: More details on JOURNAL OF SYNCHROTRON RADIATION 23, 1462-1473 (2016)
            - f(x) = norm * (gaussian_peak + charge_sharing_tail) / ((1 + α)²)
    """
    A = ((1. - alpha)**2) / (sigma * np.sqrt(2. * np.pi))
    B = 4. * alpha * (1. - alpha)
    C = 4. * alpha**2

    charge_sh_tail = (
        B - C * np.log1p((x - ped) / (q - ped))) * erfbox(x, ped, q, sigma)
    return norm * (charge_sh_tail + gauss(x, A, q, sigma)) / ((1. + alpha)**2)


def double_peak(x, ped, q_a, sigma_a, alpha, norm_a, q_b, sigma_b,  norm_b):
    """
    Model the detector response for two overlapping peaks with charge sharing.

    This function combines two charge sharing models to represent two 
    overlapping peaks in the detector response.
    """
    k_a = charge_sharing(x, ped, q_a, sigma_a, alpha, norm_a)
    k_b = charge_sharing(x, ped, q_b, sigma_b, alpha, norm_b)
    return k_a + k_b


def set_fit_range(x, x1, x2):
    """
    Set the range of x values to be used in fitting.

    This function determines the indices of x that fall within the
    specified range [x1, x2].

    Notes: If x2 <= x1, the function ensures at least two points are included
            in the range. The returned x values are cast to np.float64 for 
            higher precision in fitting.
    """
    i1 = 0
    i2 = len(x)
    i = 0

    if x2 > x1:
        if x1 > x[0]:
            while i < len(x) and x[i] < x1:
                i += 1
            i1 = i
        else:
            i1 = 0
        if x2 < x[-1]:
            while i < len(x) and x[i] < x2:
                i += 1
            i2 = i
        else:
            i2 = len(x)
    else:
        i2 = i1

    if i2 <= i1:
        i2 = i1 + 2
    # Use np.float64 for higher precision when fitting.
    # With this higher precision we avoid failed fitting
    # on very small difference.
    return x[i1:i2].astype(np.float64), i1, i2


def fit_charge_sharing(
    x, y, yerr, initial_sigma, n_sigma, ratio, alpha0=0.3
):
    """
    Perform charge sharing fit using iminuit optimization.

    This function fits the charge sharing model to the provided data using
    chi-square minimization via iminuit.

    Args:
        x (array-like): Input charge values
        y (array-like): Observed counts or intensities
        yerr (array-like): Errors on y values
        initial_sigma (float): Initial guess for peak width
        n_sigma (float): Number of sigmas for threshold calculation
            (not used in this implementation)
        ratio (float): Ratio parameter (not used in this implementation)
        alpha0: Initial guess for alpha

    Returns:
        tuple: (q, sigma, chi2ndf, alpha)
            q (float): Fitted peak position
            sigma (float): Fitted peak width
            chi2ndf (float): Chi-square per degree of freedom
            alpha (float): Fitted charge sharing probability

    Notes:
    - The pedestal (ped) is fixed to 0 in this implementation.
    - n_sigma and ratio parameters are included for compatibility
        with the original function signature
        but are not used in the current implementation.
    """
    norm = np.sum(y) * (x[1] - x[0])

    # Use _peak_position function to find initial q0
    _peaks, _ = _peak_position(x, y, thr=n_sigma*initial_sigma, ratio=ratio)
    if len(_peaks) > 0:
        q0 = np.min(x[_peaks])
    else:
        return -1, -1, -1, -1

    x_fit, i1, i2 = set_fit_range(x, 0.5 * q0, q0 + 2. * initial_sigma)

    y_fit = y[i1:i2]
    yerr_fit = yerr[i1:i2]

    def cost_function(ped, q, sigma, alpha, norm):
        # Add a small constant to yerr_fit to avoid division by zero
        epsilon = 1e-10  # Small constant to prevent division by zero
        safe_yerr = yerr_fit + epsilon
        return np.sum(
            ((y_fit - charge_sharing(
                x_fit, ped, q, sigma, alpha, norm)) / safe_yerr)**2)

    m = Minuit(
        cost_function,
        ped=0., q=q0, sigma=initial_sigma, alpha=alpha0, norm=norm,
        error_ped=0.1, error_q=0.1*q0, error_sigma=0.1*initial_sigma,
        error_alpha=0.1, error_norm=0.1*norm,
        fix_ped=True,
        limit_q=(q0 - initial_sigma, q0 + initial_sigma),
        limit_sigma=(0.1 * initial_sigma, 2. * initial_sigma),
        limit_alpha=(0.01 * alpha0, 1.),
        limit_norm=(np.sqrt(0.1 * norm), 3. * norm),
        errordef=1,  # for least squares cost function
        print_level=0,
    )

    m.migrad()

    if m.get_fmin().is_valid:
        q = m.values['q']
        sigma = m.values['sigma']
        alpha = m.values['alpha']
        chi2ndf = m.fval / (len(x_fit) - len(m.values))
    else:
        q, sigma, chi2ndf, alpha = -1000, -1000, -1000, -1000
    return q, sigma, chi2ndf, alpha


def fit_double_charge_sharing(x, y, yerr, initial_sigma, n_sigma, ratio):
    """
    Fits histogram with sum of two single photon functions with
    charge sharing tail.

    Args:
        x (array, float): x values
        y (array, float): y values
        yerr (array, float): errors of the y values
        initial_sigma (float): rough estimate of peak variance
        n_sigma (int): to calculate threshold of the peak finder as
            thr = n_sigma * initial_sigma
        ratio (float): ratio parameter of the peak finder

    Returns:
        peak position, peak variance, chi2/ndf, charge sharing parameter
        (all of them of the first peak)
    """
    alpha0 = 0.3
    norm = np.sum(y) * (x[1] - x[0])

    _peaks, _ = _peak_position(x, y, thr=n_sigma*initial_sigma, ratio=ratio)
    if len(_peaks) > 0:
        q0 = np.min(x[_peaks])
    else:
        return -1, -1, -1, -1

    q_b = 1.1 * q0
    x_fit, i1, i2 = set_fit_range(x, 0.5 * q0, q_b + initial_sigma)
    y_fit = y[i1:i2]
    yerr_fit = yerr[i1:i2]

    def cost_function(
        ped, q_a, sigma_a, alpha, norm_a, q_b, sigma_b, norm_b
    ):
        # Add a small constant to yerr_fit to avoid division by zero
        epsilon = 1e-10  # Small constant to prevent division by zero
        safe_yerr = yerr_fit + epsilon
        return np.sum((
            (y_fit - double_peak(
                x_fit, ped, q_a, sigma_a, alpha,
                norm_a, q_b, sigma_b, norm_b
                )) / safe_yerr)**2)

    limit_alpha = (0.01, 1.)
    limit_norm_a = (np.sqrt(0.1 * norm), 2. * norm)
    limit_q_b = (q_b - initial_sigma, q_b + initial_sigma)
    limit_sigma = (0.1 * initial_sigma, 2. * initial_sigma)
    limit_norm_b = (np.sqrt(0.1 * 0.2 * norm), 2 * 0.2 * norm)

    m = Minuit(
        cost_function,
        ped=0.,
        q_a=q0,
        sigma_a=initial_sigma,
        alpha=alpha0,
        norm_a=norm,
        q_b=q_b,
        sigma_b=initial_sigma,
        norm_b=norm,
        error_ped=0.1, error_q_a=0.1*q0, error_sigma_a=0.1*initial_sigma,
        error_alpha=0.1, error_norm_a=0.1*norm,
        error_q_b=0.1*q_b, error_sigma_b=0.1*initial_sigma,
        error_norm_b=0.1*norm,
        limit_q_a=(q0 - initial_sigma, q0 + initial_sigma),
        limit_sigma_a=limit_sigma,
        limit_alpha=limit_alpha,
        limit_norm_a=limit_norm_a,
        limit_q_b=limit_q_b,
        limit_sigma_b=limit_sigma,
        limit_norm_b=limit_norm_b,
        fix_ped=True,
        print_level=0,
        errordef=1,  # for least squares cost function
    )

    m.migrad()

    if m.get_fmin().is_valid:
        q = m.values['q_a']
        sigma = m.values['sigma_a']
        alpha = m.values['alpha']
        chi2ndf = m.fval / (len(x_fit) - len(m.values))
    else:
        q, sigma, chi2ndf, alpha = -1000, -1000, -1000, -1000
    return q, sigma, chi2ndf, alpha


def fit_gauss(bin_centers, histogram, yerr, initial_sigma, n_sigma, ratio):
    """
    Fits histogram with a gaussian function

    Args:
        bin_centers (array, float): bin_centers values
        histogram (array, float): histogram values
        yerr (array, float): errors of the y values
        initial_sigma (float): rough estimate of peak variance
        n_sigma (int): to calculate threshold of the peak finder as
            thr = n_sigma * initial_sigma
        ratio (float): ratio parameter of the peak finder

    Returns:
        peak position, peak variance, chi2/ndf, charge sharing parameter
        (last one alway == 0)
        all of them are kept for compatibility with the wrap around function
    """
    norm = np.sum(histogram) * (bin_centers[1] - bin_centers[0])/np.sqrt(2.*np.pi*initial_sigma**2)

    _peaks, _ = _peak_position(bin_centers, histogram, thr=n_sigma*initial_sigma, ratio=ratio)
    if len(_peaks) > 0:
        q0 = np.min(bin_centers[_peaks])
    else:
        return -1, -1, -1, -1

    x_fit, i1, i2 = set_fit_range(bin_centers, q0 - initial_sigma, q0 + 2.*initial_sigma)
    y_fit = histogram[i1:i2]
    yerr_fit = yerr[i1:i2]

    def cost_function(amp, mean, sigma):
        # Add a small constant to yerr_fit to avoid division by zero
        epsilon = 1e-10  # Small constant to prevent division by zero
        safe_yerr = yerr_fit + epsilon
        return np.sum(((
            y_fit - gauss(
                x_fit, amp, mean, sigma)) / safe_yerr)**2)

    m = Minuit(
        cost_function,
        amp=norm, mean=q0, sigma=initial_sigma,
        error_amp=0.1*norm, error_mean=0.1*initial_sigma, error_sigma=0.1*initial_sigma,
        limit_amp=(0.01*norm, 2.*norm),
        limit_mean=(q0-initial_sigma, q0+initial_sigma),
        limit_sigma=(0.1*initial_sigma, 3.*initial_sigma),
        errordef=1,  # for least squares cost function
        print_level=0,
    )

    m.migrad()

    if m.get_fmin().is_valid:
        q = m.values['mean']
        sigma = m.values['sigma']
        alpha = 0.
        chi2ndf = m.fval / (len(x_fit) - len(m.values))
    else:
        q, sigma, chi2ndf, alpha = -1000, -1000, -1000, -1000
    return q, sigma, chi2ndf, alpha
